/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-trtimeferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 */

/* Attention!
*  To maintain compliance with Nordic Semiconductor ASAs Bluetooth profile
*  qualification listings, this section of source code must not be modified.
*/
#include "ble_time_client_service.h"
#include <string.h>
#include <stdbool.h>
#include "ble_err.h"
#include "ble_srv_common.h"
#include "nordic_common.h"
#include "nrf_assert.h"
#include "device_manager.h"
#include "pstorage.h"
#include "app_error.h" 
#include "simple_uart.h"

#define START_HANDLE_DISCOVER           0x0001                                             /**< Value of start handle during discovery. */
//time data length set as 7
#define TIME_CLIENT_DATA_LENGTH        	7                                                  /**< Length of the received data. */
#define READ_DATA_LENGTH_MIN            1 
#define BLE_TIME_MAX_DISCOVERED_CENTRALS DEVICE_MANAGER_MAX_BONDS                          /**< Maximum number of discovered services that can be stored in the flash. This number should be identical to maximum number of bonded centrals. */
#define WRITE_MESSAGE_LENGTH            2 
#define TX_BUFFER_MASK                  0x07                                               /**< TX Buffer mask, must be a mask of contiguous zeroes, followed by contiguous sequence of ones: 000...111. */
#define TX_BUFFER_SIZE                 (TX_BUFFER_MASK + 1)                                /**< Size of send buffer, which is 1 higher than the mask. */
#define DISCOVERED_SERVICE_DB_SIZE \
    CEIL_DIV(sizeof(time_service_t) * BLE_TIME_MAX_DISCOVERED_CENTRALS, sizeof(uint32_t))

typedef enum
{
    READ_REQ = 1,                                                                          /**< Type identifying that this tx_message is a read request. */
    WRITE_REQ                                                                              /**< Type identifying that this tx_message is a write request. */
} time_tx_request_t;

typedef enum
{
    STATE_UNINITIALIZED,                                                                   /**< Uninitialized state of the internal state machine. */
    STATE_IDLE,                                                                            /**< Idle state, this is the state when no central has connected to this device. */
    STATE_DISC_SERV,                                                                       /**< A BLE central is connected and a service discovery is in progress. */
    STATE_DISC_CHAR,                                                                       /**< A BLE central is connected and characteristic discovery is in progress. */
    STATE_DISC_DESC,                                                                       /**< A BLE central is connected and descriptor discovery is in progress. */
    STATE_RUNNING,                                                                         /**< A BLE central is connected and complete service discovery has been performed. */
    STATE_WAITING_ENC,                                                                     /**< A previously bonded BLE central has re-connected and the service awaits the setup of an encrypted link. */
    STATE_RUNNING_NOT_DISCOVERED,                                                          /**< A BLE central is connected and a service discovery is in progress. */
} time_state_t;


/**@brief Structure used for holding the characteristic found during discovery process.
 */
typedef struct
{
    ble_uuid_t               uuid;                                                         /**< UUID identifying this characteristic. */
    ble_uuid_t							 des_uuid;
		ble_gatt_char_props_t    properties;                                                   /**< Properties for this characteristic. */
    uint16_t                 handle_decl;                                                  /**< Characteristic Declaration Handle for this characteristic. */
    uint16_t                 handle_value;                                                 /**< Value Handle for the value provided in this characteristic. */
    uint16_t                 handle_cccd;                                                  /**< CCCD Handle value for this characteristic. BLE_ANS_INVALID_HANDLE if not present in the central. */
} time_characteristic_t;

/**@brief Structure used for holding the time client Service found during discovery process.
 */
/**@brief Structure used for holding the Time stamp Service found during discovery process.
 */
typedef struct
{
    uint8_t                  handle;                                                       /**< Handle of Alert Notification Service which identifies to which central this discovered service belongs. */
    ble_gattc_service_t      service;                                                      /**< The GATT service holding the discovered Alert Notification Service. */
    time_characteristic_t    time_stamp_char;				                                       /**< Characteristic for the Alert Notification Control Point. */
} time_service_t;

/**@brief Structure for writing a message to the central, i.e. Control Point or CCCD.
 */
typedef struct
{
    uint8_t                  gattc_value[WRITE_MESSAGE_LENGTH];                            /**< The message to write. */
    ble_gattc_write_params_t gattc_params;                                                 /**< GATTC parameters for this message. */
} write_params_t;

/**@brief Structure for holding data to be transmitted to the connected central.
 */
typedef struct
{
    uint16_t                 conn_handle;                                                  /**< Connection handle to be used when transmitting this message. */
    time_tx_request_t        type;                                                         /**< Type of this message, i.e. read or write message. */
    union
    {
        uint16_t             read_handle;                                                  /**< Read request message. */
        write_params_t       write_req;                                                    /**< Write request message. */
    } req;
} tx_message_t;
static tx_message_t          m_tx_buffer[TX_BUFFER_SIZE];                                  /**< Transmit buffer for messages to be transmitted to the central. */
static uint32_t              m_tx_insert_index = 0;                                        /**< Current index in the transmit buffer where next message should be inserted. */
static uint32_t              m_tx_index = 0;                                               /**< Current index in the transmit buffer from where the next message to be transmitted resides. */
static pstorage_handle_t     m_flash_handle;                                               /**< Flash handle where discovered services for bonded masters should be stored. */
static time_state_t          m_client_state = STATE_UNINITIALIZED;                         /**< Current state of the Alert Notification State Machine. */
static uint32_t              m_service_db[DISCOVERED_SERVICE_DB_SIZE];                     /**< Service database for bonded centrals (Word size aligned). */
static time_service_t *      mp_service_db;                                                /**< Pointer to start of discovered services database. */
static time_service_t        m_service;                                                    /**< Current service data. */
static ble_time_t *          m_time;                                                  		 /**< Pointer to the instantiated object. */
static void event_connect(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
static void event_characteristic_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
static void event_discover_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
static void event_encrypted_link(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
static void event_descriptor_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
static void event_disconnect(ble_time_t * p_time);
static void event_read_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
static void tx_buffer_process(void);
static void time_pstorage_callback(pstorage_handle_t * handle,
                                  uint8_t             op_code,
                                  uint32_t            reason,
                                  uint8_t           * p_data,
                                  uint32_t            param_len);
/**@brief Function for handling of Device Manager events.
 */
void ble_time_on_device_manager_evt(ble_time_t       * p_time,
                                     dm_handle_t const * p_handle,
                                     dm_event_t const  * p_dm_evt)
{
    switch (p_dm_evt->event_id)
    {
        case DM_EVT_CONNECTION:
            // Fall through.
        case DM_EVT_SECURITY_SETUP_COMPLETE:
            p_time->central_handle = p_handle->device_id;
            break;
        default:
            // Do nothing.
            break;
    }
}
/**@brief Function for handling of BLE stack events.
 */
void ble_time_on_ble_evt(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{
    uint16_t event = p_ble_evt->header.evt_id;

        switch (m_client_state)
    {
        case STATE_UNINITIALIZED:
						 //simple_uart_putstring((const uint8_t *)" \n\rSTATE_UNINITIALIZED:\n\r ");
            // Initialization is handle in special case, thus if we enter here, it means that an
            // event is received even though we are not initialized --> ignore.
            break;

        case STATE_IDLE:
						//simple_uart_putstring((const uint8_t *)" \n\rSTATE_IDLE:\n\r ");
            if (event == BLE_GAP_EVT_CONNECTED)
            {		
						    //simple_uart_putstring((const uint8_t *)" \n\rBLE_GAP_EVT_CONNECTED:\n\r ");
                event_connect(p_time, p_ble_evt);
            }
            break;

        case STATE_WAITING_ENC:
						 //simple_uart_putstring((const uint8_t *)" \n\rSTATE_WAITING_ENC:\n\r ");
            if ((event == BLE_GAP_EVT_AUTH_STATUS) || (event == BLE_GAP_EVT_SEC_INFO_REQUEST))
            {
                event_encrypted_link(p_time, p_ble_evt);
            }
            else if (event == BLE_GAP_EVT_DISCONNECTED)
            {
                event_disconnect(p_time);
            }
            break;

        case STATE_DISC_SERV:
						//simple_uart_putstring((const uint8_t *)" \n\rSTATE_DISC_SERV:\n\r ");
            if (event == BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP)
            {
								//simple_uart_putstring((const uint8_t *)" \n\rSTATE_DISC_SERV && BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:\n\r ");
                event_discover_rsp(p_time, p_ble_evt);
            }
            else if (event == BLE_GAP_EVT_DISCONNECTED)
            {
                event_disconnect(p_time);
            }
            break;

        case STATE_DISC_CHAR:
						//simple_uart_putstring((const uint8_t *)" \n\rSTATE_DISC_CHAR:\n\r ");
            if (event == BLE_GATTC_EVT_CHAR_DISC_RSP)
            {
								//simple_uart_putstring((const uint8_t *)" \n\rSTATE_DISC_CHAR && BLE_GATTC_EVT_CHAR_DISC_RSP:\n\r ");
                event_characteristic_rsp(p_time, p_ble_evt);
								
            }
            else if (event == BLE_GAP_EVT_DISCONNECTED)
            {
                event_disconnect(p_time);
            }
            break;

        case STATE_DISC_DESC:
						//simple_uart_putstring((const uint8_t *)" \n\rSTATE_DISC_DESC:\n\r ");
            if (event == BLE_GATTC_EVT_DESC_DISC_RSP)
            {
                //simple_uart_putstring((const uint8_t *)" \n\rSTATE_DISC_DESC && BLE_GATTC_EVT_DESC_DISC_RSP :\n\r");
								event_descriptor_rsp(p_time, p_ble_evt);
            }
            else if (event == BLE_GAP_EVT_DISCONNECTED)
            {
                event_disconnect(p_time);
            }
            break;

        case STATE_RUNNING:
						//simple_uart_putstring((const uint8_t *)" \n\rSTATE_RUNNING:\n\r ");
						if (event == BLE_GATTC_EVT_READ_RSP)
            {
                //simple_uart_putstring((const uint8_t *)" \n\rBLE_GATTC_EVT_READ_RSP:\n\r ");
								event_read_rsp(p_time, p_ble_evt);
								
            }
						else if (event == BLE_GATTC_EVT_CHAR_VAL_BY_UUID_READ_RSP)
            {
                //simple_uart_putstring((const uint8_t *)" \n\rBLE_GATTC_EVT_CHAR_VAL_BY_UUID_READ_RSP:\n\r ");
								event_read_rsp(p_time, p_ble_evt);
								
            }
						else if (event == BLE_GAP_EVT_DISCONNECTED)
            {
                event_disconnect(p_time);
            }
            else
            {		
								////simple_uart_putstring((const uint8_t *)" \n\rDo nothing, event not handled in this state:\n\r ");
                // Do nothing, event not handled in this state.
            }
            break;

        case STATE_RUNNING_NOT_DISCOVERED:
            if (event == BLE_GAP_EVT_DISCONNECTED)
            {
                event_disconnect(p_time);
            }
            break;

        default:
            event_disconnect(p_time);
            break;
    }
}
/**@brief Function for initializing the time client service.
*
* @param[in]   p_time        Time client service  Service structure.
* @param[in]   p_time_init   Information needed to initialize the service.
*
* @return      NRF_SUCCESS on success, otherwise an error code.
*/

uint32_t ble_time_client_init(ble_time_t * p_time, const ble_time_init_t * p_time_init)
{
    uint32_t                err_code;
    pstorage_module_param_t param;

    if (p_time_init->evt_handler == NULL)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    p_time->evt_handler         = p_time_init->evt_handler;
    p_time->error_handler       = p_time_init->error_handler;
    p_time->service_handle      = INVALID_SERVICE_HANDLE;
    p_time->central_handle      = DM_INVALID_ID;
    p_time->message_buffer_size = p_time_init->message_buffer_size;
    p_time->p_message_buffer[0]    = p_time_init->p_message_buffer[0];
    p_time->conn_handle         = BLE_CONN_HANDLE_INVALID;

    m_time = p_time;

    memset(&m_service, 0, sizeof(time_service_t));
		memset(m_tx_buffer, 0, TX_BUFFER_SIZE);
		
    m_service.handle = INVALID_SERVICE_HANDLE;
    m_client_state   = STATE_IDLE;

    param.block_count = 1;
    param.block_size  = DISCOVERED_SERVICE_DB_SIZE * sizeof(uint32_t); // uint32_t array.
    param.cb          = time_pstorage_callback;

    // Register with storage module.
    err_code = pstorage_register(&param, &m_flash_handle);
		APP_ERROR_CHECK(err_code);
    mp_service_db = (time_service_t *) (m_service_db);

    return err_code;
}

uint32_t ble_time_client_service_load(const ble_time_t * p_time)
{
    uint32_t err_code;
    uint32_t i;

    err_code = pstorage_load((uint8_t *)m_service_db,
                             &m_flash_handle,
                             (DISCOVERED_SERVICE_DB_SIZE * sizeof(uint32_t)),
                             0);

    if (err_code != NRF_SUCCESS)
    {
        // Problem with loading values from flash, initialize the RAM DB with default.
        for (i = 0; i < BLE_TIME_MAX_DISCOVERED_CENTRALS; ++i)
        {
            mp_service_db[i].handle = INVALID_SERVICE_HANDLE;
        }

        if (err_code == NRF_ERROR_NOT_FOUND)
        {
            // The flash does not contain any memorized centrals, set the return code to success.
            err_code = NRF_SUCCESS;
        }
    }
    return err_code;
}


uint32_t ble_time_client_service_store(void)
{
    uint32_t err_code;

    err_code = pstorage_store(&m_flash_handle,
                              (uint8_t *) m_service_db,
                              DISCOVERED_SERVICE_DB_SIZE * sizeof(uint32_t),
                              0);

    return err_code;
}


uint32_t ble_time_client_service_delete(void)
{
    if (m_client_state == STATE_UNINITIALIZED)
    {
        return NRF_SUCCESS;
    }

    return pstorage_clear(&m_flash_handle, (DISCOVERED_SERVICE_DB_SIZE * sizeof(uint32_t)));
}

uint32_t ble_time_alert_read(const ble_time_t * p_time)
{
    tx_message_t * msg;

    msg                = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index &= TX_BUFFER_MASK;

    msg->req.read_handle = m_service.time_stamp_char.handle_value;
		//msg->req.read_handle	 =m_service.time_stamp_char.handle_decl;
    msg->conn_handle     = p_time->conn_handle;
    msg->type            = READ_REQ;
	
    tx_buffer_process();
    return NRF_SUCCESS;
}
uint32_t ble_time_cccd_enable(const ble_time_t * p_time)
{
    return cccd_configure(p_time->conn_handle,
                          m_service.time_stamp_char.handle_cccd,
                          true);
}

/**@brief Function for creating a TX message for writing a CCCD.
 */
static uint32_t cccd_configure(uint16_t conn_handle, uint16_t handle_cccd, bool enable)
{
    tx_message_t * p_msg;
    uint16_t       cccd_val = enable ? BLE_GATT_HVX_NOTIFICATION : 0;

    if (m_client_state != STATE_RUNNING)
    {
        return NRF_ERROR_INVALID_STATE;
    }

    p_msg              = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index &= TX_BUFFER_MASK;

    p_msg->req.write_req.gattc_params.handle   = handle_cccd;
    p_msg->req.write_req.gattc_params.len      = WRITE_MESSAGE_LENGTH;
    p_msg->req.write_req.gattc_params.p_value  = p_msg->req.write_req.gattc_value;
    p_msg->req.write_req.gattc_params.offset   = 0;
    p_msg->req.write_req.gattc_params.write_op = BLE_GATT_OP_WRITE_REQ;
    p_msg->req.write_req.gattc_value[0]        = LSB(cccd_val);
    p_msg->req.write_req.gattc_value[1]        = MSB(cccd_val);
    p_msg->conn_handle                         = conn_handle;
    p_msg->type                                = WRITE_REQ;

    tx_buffer_process();
    return NRF_SUCCESS;
}

/**@brief Function for passing any pending request from the buffer to the stack.
 */

static void tx_buffer_process(void)
{
    if (m_tx_index != m_tx_insert_index)
    {
        uint32_t err_code;

        if (m_tx_buffer[m_tx_index].type == READ_REQ)
        {
            err_code = sd_ble_gattc_read(m_tx_buffer[m_tx_index].conn_handle,
                                         m_tx_buffer[m_tx_index].req.read_handle,
                                         0);
						if (err_code == NRF_SUCCESS){simple_uart_putstring((const uint8_t *)" \n\rread success:\n\r ");}
						else{simple_uart_putstring((const uint8_t *)" \n\rread request failed:\n\r ");}
				}
        else
        {
            err_code = sd_ble_gattc_write(m_tx_buffer[m_tx_index].conn_handle,
                                          &m_tx_buffer[m_tx_index].req.write_req.gattc_params);
						if (err_code == NRF_SUCCESS){simple_uart_putstring((const uint8_t *)" \n\rwrite success:\n\r ");}
						else {simple_uart_putstring((const uint8_t *)" \n\rwrite request failed:\n\r ");}
        }
        if (err_code == NRF_SUCCESS)
        {		
            ++m_tx_index;
            m_tx_index &= TX_BUFFER_MASK;
        }
    }
}
/**@brief Function for validating and passing the response to the application,
 *			when a read response is received.
 */

static void event_read_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{
    ble_time_evt_t                  event;  
    const ble_gattc_evt_read_rsp_t * p_response;
		simple_uart_putstring((const uint8_t *)" \n\rinside the event read response:\n\r ");

    p_response     = &p_ble_evt->evt.gattc_evt.params.read_rsp;
	
    if (p_response->len < READ_DATA_LENGTH_MIN)
    {		
				simple_uart_putstring((const uint8_t *)" \n\rread response data len less the data length minimum:\n\r ");
        tx_buffer_process();
        return;
    }

    if (p_response->handle == m_service.time_stamp_char.handle_value)								// == m_service.chars.handle_value)
    {		
				simple_uart_putstring((const uint8_t *)" \n\rresponse handle equal to the time handle:\n\r ");
        BLE_UUID_COPY_INST(event.evt_chars.uuid, m_service.time_stamp_char.uuid);
				event.evt_attr_len=p_response->len;
				event.evt_attr_data[0]=p_response->data[0];
				event.evt_attr_data[1]=p_response->data[1];
				event.evt_attr_data[2]=p_response->data[2];
				event.evt_attr_data[3]=p_response->data[3];
				event.evt_attr_data[4]=p_response->data[4];
			
				event.evt_type = BLE_TIME_EVT_READ_RESP;
				p_time->evt_handler(&event);
    }
    else
				
		{		simple_uart_putstring((const uint8_t *)" \n\rbad response:\n\r ");
        // Bad response, ignore.
        tx_buffer_process();
        return;
    }
		
    //event.data.settings = *(ble_time_alert_settings_t *)(p_response->data);
		//ble_gattc_handle_range_t char_handle_range;
		
    

    tx_buffer_process();
}

/**@brief Function for disconnecting and cleaning the current service.
 */

static void event_disconnect(ble_time_t * p_time)
{

    m_client_state = STATE_IDLE;

    if (m_service.handle == INVALID_SERVICE_HANDLE_DISC &&
        p_time->central_handle != DM_INVALID_ID)
    {
        m_service.handle = p_time->central_handle;
    }

    if (m_service.handle < BLE_TIME_MAX_DISCOVERED_CENTRALS)
    {
        mp_service_db[m_service.handle] = m_service;
    }

    memset(&m_service, 0, sizeof(time_service_t));

    m_service.handle       = INVALID_SERVICE_HANDLE;
    p_time->service_handle = INVALID_SERVICE_HANDLE;
    p_time->conn_handle    = BLE_CONN_HANDLE_INVALID;
    p_time->central_handle = DM_INVALID_ID;
}
/**@brief Function for setting the discovered descriptor in the alert service.
 */
static void descriptor_set(time_service_t * p_service, const ble_gattc_desc_t * p_desc_resp)
{		
		simple_uart_putstring((const uint8_t *)" \n\rinside the descriptor_set: ");
    if (p_service->time_stamp_char.handle_value == (p_desc_resp->handle - 1) &&
        p_desc_resp->uuid.uuid == BLE_UUID_DESCRIPTOR_CLIENT_CHAR_CONFIG)
    {
        p_service->time_stamp_char.handle_cccd = p_desc_resp->handle;
				p_service->time_stamp_char.des_uuid=p_desc_resp->uuid;
				
				simple_uart_putstring((const uint8_t *)" \n\rinside the descriptor_set if condition satisfied: ");
    }
}
/**@brief Function for updating the current state and sending an event on discovery failure.
 */
static void handle_discovery_failure(const ble_time_t * p_time, uint32_t code)
{
    m_client_state        = STATE_RUNNING_NOT_DISCOVERED;
}
/**@brief Function for executing the Characteristic Descriptor Discovery Procedure.
 */
static void descriptor_disc_req_send(const ble_time_t * p_time)
{
    ble_gattc_handle_range_t descriptor_handle;
    uint32_t                 err_code = NRF_SUCCESS;
		simple_uart_putstring((const uint8_t *)" \n\rinside the descriptor_disc_req_send: ");
    // If we have not discovered descriptors for all characteristics,
    // we will discover next descriptor.
    if (m_service.time_stamp_char.handle_cccd == BLE_TIME_INVALID_HANDLE)
    {
        descriptor_handle.start_handle = m_service.time_stamp_char.handle_value + 1;
        descriptor_handle.end_handle   = m_service.time_stamp_char.handle_value + 1;

        err_code = sd_ble_gattc_descriptors_discover(p_time->conn_handle, &descriptor_handle);
    }

    if (err_code == NRF_SUCCESS)
    {
        m_client_state = STATE_DISC_DESC;
				simple_uart_putstring((const uint8_t *)" \n\rdescriptor_disc_req_send success: ");
    }
    else
    {
        handle_discovery_failure(p_time, err_code);
    }
}
/**@brief Function for indicating that a connection has successfully been established.
 *        Either when the Service Discovery Procedure completes or a re-connection has been
 *        established to a bonded central.
 *
 * @details This function is executed when a service discovery or a re-connect to a bonded central
 *          occurs. In the event of re-connection to a bonded central, this function will ensure
 *          writing of the control point according to the Alert Notification Service Client
 *          specification.
 *          Finally an event is passed to the application:
 *          BLE_ANS_C_EVT_RECONNECT         - When we are connected to an existing central and the
 *                                            alert notification control point has been written.
 *          BLE_ANS_C_EVT_DISCOVER_COMPLETE - When we are connected to a new central and the Service
 *                                            Discovery has been completed.
 */
static void connection_established(const ble_time_t * p_time)
{
		ble_time_evt_t event;
		simple_uart_putstring((const uint8_t *)" \n\rconnection_established function m_client_state = STATE_RUNNING: ");
    m_client_state = STATE_RUNNING;

    if (m_service.handle == INVALID_SERVICE_HANDLE)
    {
        m_service.handle = INVALID_SERVICE_HANDLE_DISC;
    }
		if(p_time->central_handle != DM_INVALID_ID &&
        m_service.handle < INVALID_SERVICE_HANDLE_BASE)
		{
					simple_uart_putstring((const uint8_t *)" \n\rconnection_established function for notify request: ");
		}
		else
		{		
        event.evt_type = BLE_TIME_EVT_DISCOVER_COMPLETED;
        p_time->evt_handler(&event);
		}
}
/**@brief Function for handling of descriptor discovery responses.
 *
 * @details This function will validate and store the descriptor received.
 *          If not all descriptors are discovered it will continue the descriptor discovery
 *          procedure, otherwise it will continue to running state.
 *          If we receive a GATT Client error, we will go to handling of discovery failure.
 */

static void event_descriptor_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{
    if (p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND ||
        p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_ATTERR_INVALID_HANDLE)
    {		
				simple_uart_putstring((const uint8_t *)" \n\rhandle_discovery_failure: ");
        handle_discovery_failure(p_time, NRF_ERROR_NOT_FOUND);
    }
    else if (p_ble_evt->evt.gattc_evt.gatt_status)
    {
        // We have received an unexpected result.
        // As we are in a connected state, but can not continue
        // our descriptor discovery, we will go to running state.
				//simple_uart_putstring((const uint8_t *)" \n\rinside the else if condition: ");
        handle_discovery_failure(p_time, p_ble_evt->evt.gattc_evt.gatt_status);
    }
    else
    {
				simple_uart_putstring((const uint8_t *)" \n\rinside the else condition: ");
        if (p_ble_evt->evt.gattc_evt.params.desc_disc_rsp.count > 0)
        {
						simple_uart_putstring((const uint8_t *)" \n\rdescriptor_set: ");
            descriptor_set(&m_service, &(p_ble_evt->evt.gattc_evt.params.desc_disc_rsp.descs[0]));
        }
        if (m_service.time_stamp_char.handle_cccd == BLE_TIME_INVALID_HANDLE)
        {
            descriptor_disc_req_send(p_time);
        }
        else
        {
            connection_established(p_time);
        }
    }
}



/**@brief Function for executing the Service Discovery Procedure.
 */
static void service_disc_req_send(const ble_time_t * p_time)
{
    uint16_t   handle = START_HANDLE_DISCOVER;
    ble_uuid_t time_uuid;
    uint32_t   err_code;

    // Discover services on uuid current time service.
    BLE_UUID_BLE_ASSIGN(time_uuid, BLE_UUID_CURRENT_TIME_SERVICE);

    err_code = sd_ble_gattc_primary_services_discover(p_time->conn_handle, handle, &time_uuid);
    if (err_code != NRF_SUCCESS)
    {
        handle_discovery_failure(p_time, err_code);
    }
    else
    {
        m_client_state = STATE_DISC_SERV;
    }
}

/**@brief Function for executing the Characteristic Discovery Procedure.
 */
static void characteristic_disc_req_send(const ble_time_t              * p_time,
                                         const ble_gattc_handle_range_t * p_handle)
{
    uint32_t err_code;

    // With valid service, we should discover characteristics.
    err_code = sd_ble_gattc_characteristics_discover(p_time->conn_handle, p_handle);

    if (err_code == NRF_SUCCESS)
    {
        m_client_state = STATE_DISC_CHAR;
    }
    else
    {
        handle_discovery_failure(p_time, err_code);
    }
}


/**@brief Function for waiting until an encrypted link has been established to a bonded central.
 */
static void encrypted_link_setup_wait(const ble_time_t * p_ans)
{
    m_client_state = STATE_WAITING_ENC;
}

/**@brief Function for handling the connect event when a central connects.
 *
 * @details This function will check if bonded central connects, and do the following
 *          Bonded central  - enter wait for encryption state.
 *          Unknown central - Initiate service discovery procedure.
 */
static void event_connect(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{
    p_time->conn_handle = p_ble_evt->evt.gatts_evt.conn_handle;

    if (p_time->central_handle != DM_INVALID_ID)
    {
        m_service = mp_service_db[p_time->central_handle];
        encrypted_link_setup_wait(p_time);
    }
    else
    {
        m_service.handle = INVALID_SERVICE_HANDLE;
        service_disc_req_send(p_time);
    }
}

/**@brief Function for handling the encrypted link event when a secure
 *        connection has been established with a central.
 *
 * @details This function will check if the service for the central is known.
 *          Service known   - Execute action running / switch to running state.
 *          Service unknown - Initiate Service Discovery Procedure.
 */

static void event_encrypted_link(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{
    // If we are setting up a bonded connection and the UUID of the service
    // is unknown, a new discovery must be performed.
    if (m_service.service.uuid.uuid != BLE_UUID_CURRENT_TIME_SERVICE)
    {
        m_service.handle = INVALID_SERVICE_HANDLE;
        service_disc_req_send(p_time);
    }
    else
    {
        connection_established(p_time);
    }
}
static void time_pstorage_callback(pstorage_handle_t * handle,
                                  uint8_t             op_code,
                                  uint32_t            reason,
                                  uint8_t           * p_data,
                                  uint32_t            param_len)
{
    if (reason != NRF_SUCCESS)
    {
        m_time->error_handler(reason);
    }
}
/**@brief Function for handling the response on service discovery.
 *
 * @details This function will validate the response.
 *          Invalid response - Handle discovery failure.
 *          Valid response   - Initiate Characteristic Discovery Procedure.
 */

static void event_discover_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{
    if (p_ble_evt->evt.gattc_evt.gatt_status != BLE_GATT_STATUS_SUCCESS)
    {
        // We have received an  unexpected result.
        // As we are in a connected state, but can not continue
        // our service discovery, we will go to running state.
        handle_discovery_failure(p_time, p_ble_evt->evt.gattc_evt.gatt_status);
    }
    else
    {
        BLE_UUID_COPY_INST(m_service.service.uuid,
                           p_ble_evt->evt.gattc_evt.params.prim_srvc_disc_rsp.services[0].uuid);

        if (p_ble_evt->evt.gattc_evt.params.prim_srvc_disc_rsp.count > 0)
        {
            const ble_gattc_service_t * p_service;

            p_service = &(p_ble_evt->evt.gattc_evt.params.prim_srvc_disc_rsp.services[0]);

            m_service.service.handle_range.start_handle = p_service->handle_range.start_handle;
            m_service.service.handle_range.end_handle   = p_service->handle_range.end_handle;

            characteristic_disc_req_send(p_time, &(m_service.service.handle_range));
        }
        else
        {
            // If we receive an answer, but it contains no service, we just go to state: Running.
            handle_discovery_failure(p_time, p_ble_evt->evt.gattc_evt.gatt_status);
        }
    }
}
/**@brief Function for setting the discovered characteristics in the alert service.
 */
static void characteristics_set(time_characteristic_t * p_characteristic,
                                const ble_gattc_char_t * p_char_resp)
{		simple_uart_putstring((const uint8_t *)" \n\rinside the characteristics set: ");
    BLE_UUID_COPY_INST(p_characteristic->uuid, p_char_resp->uuid);

    p_characteristic->properties   = p_char_resp->char_props;
    p_characteristic->handle_decl  = p_char_resp->handle_decl;
    p_characteristic->handle_value = p_char_resp->handle_value;
		p_characteristic->handle_cccd  = BLE_TIME_INVALID_HANDLE;
}
/**@brief Function for handling a characteristic discovery response event.
 *
 * @details This function will validate and store the characteristics received.
 *          If not all characteristics are discovered it will continue the characteristic discovery
 *          procedure, otherwise it will continue with the descriptor discovery procedure.
 *          If we receive a GATT Client error, we will go to handling of discovery failure.
 */
static void event_characteristic_rsp(ble_time_t * p_time, const ble_evt_t * p_ble_evt)
{		
		simple_uart_putstring((const uint8_t *)" \n\rinside event_characteristic_rsp function:\n\r ");
    if (p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND ||
        p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_ATTERR_INVALID_HANDLE)
    {		
        if (m_service.time_stamp_char.handle_value == 0)
        {
            // At least one required characteristic is missing on the server side.
						//simple_uart_putstring((const uint8_t *)" \n\rhandle_discovery_failure: ");
						simple_uart_putstring((const uint8_t *)" \n\rhandle_discovery_failure: ");
            handle_discovery_failure(p_time, NRF_ERROR_NOT_FOUND);
						
        }
        else
        {
						simple_uart_putstring((const uint8_t *)" \n\rdescriptor_disc_req_send from event_characteristic_rsp: ");
            descriptor_disc_req_send(p_time);
        }
    }
    else if (p_ble_evt->evt.gattc_evt.gatt_status)
    {
        simple_uart_putstring((const uint8_t *)" \n\rhandle_discovery_failure: ");
				// We have received an  unexpected result.
        // As we are in a connected state, but can not continue
        // our service discovery, we must go to running state.
        handle_discovery_failure(p_time, p_ble_evt->evt.gattc_evt.gatt_status);
    }
    else
    {
    {
        uint32_t                 i;
        const ble_gattc_char_t * p_char_resp = NULL;

        // Iterate trough the characteristics and find the correct one.
        for (i = 0; i < p_ble_evt->evt.gattc_evt.params.char_disc_rsp.count; i++)
        {
            p_char_resp = &(p_ble_evt->evt.gattc_evt.params.char_disc_rsp.chars[i]);
            switch (p_char_resp->uuid.uuid)
            {
                case BLE_UUID_CURRENT_TIME_CHAR:
										simple_uart_putstring((const uint8_t *)" \n\rcase BLE_UUID_CURRENT_TIME_CHAR: ");
                    characteristics_set(&m_service.time_stamp_char, p_char_resp);
                    break;

                default:
                    // No implementation needed.
                    break;
            }
        }

        if (p_char_resp != NULL)
        {		
						simple_uart_putstring((const uint8_t *)" \n\rif (p_char_resp != NULL): ");
            // If not all characteristics are received, we do a 2nd/3rd/... round.
            ble_gattc_handle_range_t char_handle;

            char_handle.start_handle = p_char_resp->handle_value + 1;
            char_handle.end_handle   = m_service.service.handle_range.end_handle;

            characteristic_disc_req_send(p_time, &char_handle);
        }
        else
        {
            characteristic_disc_req_send(p_time, &(m_service.service.handle_range));
        }
    } 
}
}