

#ifndef BLE_TIME_CLIENT_H__
#define BLE_TIME_CLIENT_H__

#include "ble.h"
#include "ble_gatts.h"
#include "ble_types.h"
#include "ble_srv_common.h"
#include "device_manager.h"

#define TIME_NB_OF_CHARACTERISTICS                   1                                     /**< Number of characteristics as defined by Alert Notification Service specification. */
#define TIME_NB_OF_SERVICES                          1                                     /**< Number of services supported in one central. */
#define INVALID_SERVICE_HANDLE_BASE                 0xF0                                  /**< Base for indicating invalid service handle. */
#define INVALID_SERVICE_HANDLE                      (INVALID_SERVICE_HANDLE_BASE + 0x0F)  /**< Indication that the current service handle is invalid. */
#define INVALID_SERVICE_HANDLE_DISC                 (INVALID_SERVICE_HANDLE_BASE + 0x0E)  /**< Indication that the current service handle is invalid but the service has been discovered. */
#define BLE_TIME_INVALID_HANDLE                      0xFF                                  /**< Indication that the current service handle is invalid. */

// Forward declaration of the ble_time_t type.
typedef struct ble_time_s ble_time_t;

/**@brief Time server Event types that are passed from client to application on an event. */
typedef enum
{
    BLE_TIME_EVT_DISCOVER_COMPLETED,
		BLE_TIME_EVT_NOTIFICATION,                                                           /**< A valid Alert Notification has been received from the server.*/
		BLE_TIME_EVT_READ_RESP
} ble_time_evt_type_t;

//typedef struct
//{
//    uint32_t                            time_stamp_length;                                  /**< Length of optional text message send by the server. */
//    uint8_t *                           p_time_stamp_buf;                                   /**< Pointer to buffer containing the optional text message. */
//} ble_time_stamp_t;

/**@brief time stamp Event structure
 *
 * @details The structure contains the event that should be handled, as well as
 *          additional information.
 */
typedef struct
{		ble_time_evt_type_t					evt_type;
		uint8_t                  		evt_conn_handle;                                                      	 	/**< Handle of time client Service which identifies to which central this discovered service belongs. */
		uint16_t            		 		evt_count;           																											/**< Service count. */
		ble_gattc_service_t 				evt_service; 		     																											/**< Service data, variable length. */
		ble_gattc_char_t						evt_chars;
		ble_gattc_desc_t						evt_desc;
		uint16_t            				evt_attr_handle;
		uint16_t            				evt_attr_offset;
		uint16_t            				evt_attr_len;
		uint8_t   		        			evt_attr_data[7];
} ble_time_evt_t;


/**@brief Alert Notification event handler type. */
typedef void (*ble_time_evt_handler_t) (ble_time_evt_t * p_evt);
/**@brief Alert Notification structure. This contains various status information for the client. */
typedef struct ble_time_s
{
    ble_time_evt_handler_t              evt_handler;                                      /**< Event handler to be called for handling events in the Alert Notification Client Application. */
    ble_srv_error_handler_t             error_handler;                                    /**< Function to be called in case of an error. */
    uint16_t                            conn_handle;                                      /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection). */
    uint8_t                             central_handle;                                   /**< Handle for the currently connected central if peer is bonded. */
    uint8_t                             service_handle;                                   /**< Handle to the service in the database to use for this instance. */
    uint8_t                            	message_buffer_size;                              /**< Size of message buffer to hold the additional text messages received on notifications. */
    uint8_t                            	p_message_buffer[7];                                 /**< Pointer to the buffer to be used for additional text message handling. */
} ble_time_t;

/**@brief Alert Notification init structure. This contains all options and data needed for
 *        initialization of the client.*/
typedef struct
{
    ble_time_evt_handler_t              evt_handler;                                      /**< Event handler to be called for handling events in the Battery Service. */
    ble_srv_error_handler_t             error_handler;                                    /**< Function to be called in case of an error. */
    uint32_t                            message_buffer_size;                              /**< Size of buffer to handle messages. */
    uint8_t *                           p_message_buffer;                                 /**< Pointer to buffer for passing messages. */
} ble_time_init_t;

/**@brief Function for initializing the time client service.
*
* @param[in]   p_time        Time client service  Service structure.
* @param[in]   p_time_init   Information needed to initialize the service.
*
* @return      NRF_SUCCESS on success, otherwise an error code.
*/

uint32_t ble_time_client_init(ble_time_t * p_time, const ble_time_init_t * p_time_init);

/**@brief Function for handling of BLE stack events.
 */
void ble_time_on_ble_evt(ble_time_t * p_time, const ble_evt_t * p_ble_evt);
/**@brief Function for handling of Device Manager events.
 */
void ble_time_on_device_manager_evt(ble_time_t       * p_time,
                                     dm_handle_t const * p_handle,
                                     dm_event_t const  * p_dm_evt);

/**@brief  Function for loading previous discovered service and characteristic handles for bonded centrals from
 *          flash into RAM.
 *
 * @details Read the database of all discovered service and characteristic handles from flash.
 *          If the flash does not contain any valid data, the array of discovered service handles in
 *          RAM will be empty.
 *
 * @param[in] p_ans  Alert Notification structure. This structure will have to be supplied by the
 *                   application. It identifies the particular client instance to use.
 *
 * @note    Currently the Alert Notification Client uses only one page in flash.
 *
 * @return  NRF_SUCCESS if all operations went successfully, an error_code otherwise.
 */
uint32_t ble_time_client_service_load(const ble_time_t * p_time);


/**@brief Function for storing discovered service and characteristic handles for bonded centrals into flash memory.
 *
 * @details This function will erase the flash page (if the data to store
 *          are diferent than the one already stored) and then write into flash. Those
 *          operations could prevent the radio to run.
 *
 * @note    Do not call this function while in a connection or when advertising. If you do, the
 *          behavior is undefined.
 *
 * @return  NRF_SUCCESS if all operations went successfully, an error_code otherwise.
 */
uint32_t ble_time_client_service_store(void);


/**@brief Function for deleting the Alert Notification Client database from flash.
 *
 * @details After calling this function you should call ble_ans_c_init(...) to re-initialize
 *          the RAM database.
 *
 * @return  NRF_SUCCESS if all operations went successfully.
 */
uint32_t ble_time_client_service_delete(void);
uint32_t ble_time_alert_read(const ble_time_t * p_time);
static uint32_t cccd_configure(uint16_t conn_handle, uint16_t handle_cccd, bool enable);
uint32_t ble_time_cccd_enable(const ble_time_t * p_time);


#endif //BLE_TIME_CLIENT_H__





















































